package de.hendriklipka.config.impl;

import org.apache.xerces.parsers.DOMParser;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.util.Date;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import de.hendriklipka.config.impl.BlockImpl;
import de.hendriklipka.config.impl.BlockReplacementHandler;
import de.hendriklipka.config.*;

/**
 * @todo handle reload changes (notify via observers)
 * @todo documentation (code and javadoc)
 * @todo error handling
 */
public class ConfigurationImpl implements Configuration
{
    /**
     *  list of all input files read by the configuration
     */
    protected Vector _inputFiles=new Vector();
    /**
     * name and path of root config file
     */
    protected String _configFile=null;
    /**
     * name of config file
     */
    private final static String CONFIGFILENAME="config.xml";

    /**
     *  root block
     */
    protected BlockImpl _root=null;

    private final static ConfigurationImpl _instance=new ConfigurationImpl();

    protected Vector _handledBlocks=null;
    protected Vector _unhandledBlocks=null;

    protected Vector _replacementHandlers=new Vector();

    public void addSourceFile(String fileName)
    {
        _inputFiles.addElement(fileName);
    }

    /**
     *  Gets the RootBlock of the ConfigurationImpl object
     *
     *@return    The RootBlock value
     *
     * @post $return!=null
     */
    public Block getRootBlock()
    {
        return _root;
    }

    /**
     *  returns the block specified by the given name. Specifying sub-blocks (with
     *  "super.sub.block") is possible.
     *
     *@param  name  the block name
     *@return       the block
     *
     * @pre name!=null
     */
    public Block getBlockByName(String name)
    {
        return getRootBlock().getChild(name);
    }

    /**
     *  returns all blocks specified by the given name. Specifying sub-blocks (with
     *  "super.sub.block") is possible.
     *
     *@param  name  the name
     *@return       enumeration containing all block matching the name
     *
     * @pre name!=null
     */
    public Enumeration getBlocksByName(String name)
    {
        return getRootBlock().getChilds(name);
    }

    /**
     *  initialize the ConfigurationImpl object
     *
     */
    public static Configuration createConfigFromFile(String configFile) throws ConfigFileException
    {
        ConfigurationImpl config=new ConfigurationImpl();
        config.addReplacementHandler(new BlockReplacementHandler());
        config._configFile=configFile;
        config._root=config.readConfigurationFromFile();
        Thread t=new Thread(new FileWatcherThread(config, config._inputFiles), "FileWatcherThread");
        t.setDaemon(true);
        t.start();
        return config;
    }

    public static Configuration createConfigurationFromNode(Node node) throws ConfigFileException
    {
        ConfigurationImpl config=new ConfigurationImpl();
        config._root=new BlockImpl();
        config._root.readConfigurationFromXML(node,null,null);
        return config;
    }

    /**
     *  reloads the configuration
     */
    public void reload() throws ConfigFileException
    {
        BlockImpl root=readConfigurationFromFile();
        if (null!=root)
        {
//            handleReplacement(root);
            _root=root;
        }
        else
            System.out.println("error while reloading");
    }

    /**
     *  returns the singleton instance of the COnfiguration class
     *
     *@return    the instance
     *
     * @post $return !=null
     */
    public static ConfigurationImpl getInstance()
    {
        return _instance;
    }

    /**
     * reads the configuration from the given file
     *
     *@todo             exception handling
     *@todo handle relative filename
     *
     * @pre fileName!=null
     * @pre true==(new File(fileName).exists())
     */
    protected BlockImpl readConfigurationFromFile() throws ConfigFileException
    {
        String fileName=_configFile;
        
        try
        {
            _inputFiles.addElement(fileName);
            String baseDir=new File(fileName).getParentFile().getAbsolutePath();
            InputStream result=new FileInputStream(fileName);
            return readConfiguration(result, baseDir);
        }
        catch (ConfigFileException ex)
        {
            throw ex;
        }
        catch (Exception ex)
        {
            throw new ConfigFileException(ex.getMessage());
        }
    }

    /**
     *  reads the configuration from the given input string
     *
     *@param  inStream          the InputStream for the config
     *
     * @pre inStream!=null
     */
    protected BlockImpl readConfiguration(InputStream inStream, String baseDir) throws ConfigFileException
    {
        DOMParser parser=new DOMParser();
        try
        {
            parser.parse(new InputSource(inStream));
            Document doc=parser.getDocument();
            Element el=doc.getDocumentElement();
            BlockImpl root=new BlockImpl();
            root.readConfigurationFromXML(el, baseDir, null);
            return root;
        }
        catch (ConfigFileException ex)
        {
            throw ex;
        }
        catch (Exception e)
        {
            e.printStackTrace();
            throw new ConfigFileException(e.getMessage());
        }
    }
    
    public boolean isAlreadyRead(String fileName)
    {
        File f=new File(fileName);
        for (int i=0; i<_inputFiles.size(); i++)
        {
            String fName=(String) _inputFiles.elementAt(i);
            if (0==(new File(fName).compareTo(f)))
                return true;
        }
        return false;
    }

    public void addUnhandledBlock(Block block)
    {
        _unhandledBlocks.addElement(block);
    }

    public void addHandledBlock(Block block)
    {
        _handledBlocks.addElement(block);
    }

    protected static class FileWatcherThread implements Runnable
    {
        protected Vector _files=null;
        protected Hashtable _fileDates=new Hashtable();
        private Configuration _config;

        public FileWatcherThread(Configuration config, Vector files)
        {
            _config=config;
            _files=files;
        }

        public void run()
        {
            while (true)
            {
                for (int i=0; i<_files.size(); i++)
                {
                    boolean r=checkFileForChange(i);
                    if (r)
                    {
                        try
                        {
                            _config.reload();
                        }
                        catch (ConfigFileException e)
                        {
                            e.printStackTrace();
                        }
                        break;
                    }
                }
                synchronized (this)
                {
                    try
                    {
                        wait(5*1000);
                    }
                    catch (InterruptedException e)
                    {
                    }
                }
            }
        }

        private boolean checkFileForChange(int i)
        {
            String fileName=(String) _files.elementAt(i);
            File file=new File(fileName);
            if (!file.exists())
            {
                _files.removeElementAt(i);
                return true;
            }
            Date currentDate=new Date(file.lastModified());
            Date lastDate=(Date) _fileDates.get(fileName);
            if (null==lastDate)
            {
                _fileDates.put(fileName, currentDate);
            }
            else
            {
                if (!lastDate.equals(currentDate))
                {
                    _fileDates=new Hashtable();
                }
            }
            return false;
        }
    }

    public void addReplacementHandler(ReplacementHandler handler)
    {
        _replacementHandlers.addElement(handler);
    }

    Enumeration getReplacementHandlers()
    {
        return _replacementHandlers.elements();
    }

    public static void main(String[] args)
    {
        String configFile="."+File.separator+"cfg"+File.separator+CONFIGFILENAME;
        try
        {
            ConfigurationImpl.getInstance().createConfigFromFile(configFile);
        }
        catch (ConfigFileException e)
        {
            e.printStackTrace();
        }
        Block b=ConfigurationImpl.getInstance().getRootBlock();
        System.out.println(b.toString());
/*
        BlockImpl b2=b.getChild("block4.block41");
        System.out.println(b2.getValue());
        Enumeration e=b.getChilds("block4");
        while (e.hasMoreElements())
        {
            BlockImpl block=(BlockImpl)e.nextElement();
            Enumeration e2=block.getChilds("block41");
            while (e2.hasMoreElements())
            {
                BlockImpl block2=(BlockImpl)e2.nextElement();
                System.out.println(block2.getValue());
            }
        }
*/
    }
}
