/*
Copyright (C) 1999/2000 Hendrik Lipka.
All rights reserved.

This program is free software; you can redistribute it and/or modify it under the terms 
of the GNU General Public License as published by the Free Software Foundation; 
either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License 
for more details. 

You should have received a copy of the GNU General Public License along with this 
program; if not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
Cambridge, MA 02139, USA or see http://www.gnu.org

If you encounter problems or have a suggestion, please mail to:
hli@hendriklipka.de
*/

package de.hendriklipka.util;

import java.util.*;
import java.lang.reflect.*;

import org.w3c.dom.*;
import de.hendriklipka.config.Block;

public class XMLUtil
{
    // extracts the string content from a XML node
    public static String getTextData(Node node)
    {
        NodeList subs=node.getChildNodes();
        String s="";
        for (int j=0;j<subs.getLength();j++)
        {
            Node attr=subs.item(j);
            if ((attr.getNodeType()!=attr.TEXT_NODE)&&(attr.CDATA_SECTION_NODE!=attr.getNodeType()))
                continue;
            Text val=(Text)attr;
            s+=val.getData().trim();
        }
        if (0==s.length())
            return null;
        return s;
    }
    
    /**
        reads a XML node, parses it and writes the values according to the rules
        in a destination object
        attributes in the destination object must be public

        rules:
        2-dimensional string array
        {{"var name","element name"}, // writes the content of this element into the variable 
         {"var name","#attrib name"}, // writes the content of the attribute into the variable
         {"method name()","element name"}} // calls the method with the element
                                         // parameters are the node and a properties opbject (funcData)
               
        checking rules:
        set of strings, in RPN:
        "var1" // var1 must have assigned a value
        "var2 var3 ||" // var3 and var2 must have assigned a value
        "var4 var5 && var6 var7 && ||" // var4 and var5 OR var6 and var7 must have assigned a value
        var means a variable name or a method name, which get assigned a value (or was called)
        each rule must me true after a boolean evaluation
        operators: && || !
        
        @param configNode the node with the configuration data
        @param rules the mapping rules
        @param checkRules the checking rules
        @param funcData the data which is given to each called method
        @param destination the destination object
    */
    public static void readXMLConfigNode(Node configNode,String[][] rules, String[] checkRules, Properties funcData, Object destination) throws Exception
    { 
        Hashtable attributes=new Hashtable();
        Hashtable elements=new Hashtable();
        Hashtable found=new Hashtable();
        for (int ruleNum=0;ruleNum<rules.length;ruleNum++)
        {
            try
            {
                String[] rule=rules[ruleNum];
                String dest=rule[0];
                String name=rule[1];
                if ('#'==name.charAt(0))
                {
                    attributes.put(name,dest);
                }
                else
                {
                    elements.put(name,dest);
                }
            }
            catch (StringIndexOutOfBoundsException ex)
            {
                ex.printStackTrace();
                throw new Exception("Rule "+ruleNum+" has not enough elements");
            }
        }
        // attributes
        NamedNodeMap nnm=configNode.getAttributes();
        for (int i=0;i<nnm.getLength();i++)
        {
            Node node=nnm.item(i);
            String name=node.getNodeName();
            Object o=elements.get(name);
            if (null!=o)
            {
                try
                {
                    String data=node.getNodeValue();
                    if (putVariable(destination,(String)o,data))
                        found.put(o,o);
                }
                catch (Exception ex)
                {
                    ex.printStackTrace();
                    throw new Exception("error while trying to set "+name+": "+ex.getMessage());
                }
            }
        }
        // sub-nodes
        NodeList nl=configNode.getChildNodes();
        for (int i=0;i<nl.getLength();i++)
        {
            Node node=nl.item(i);
            if (node.getNodeType()==node.ELEMENT_NODE)
            {
                String name=node.getNodeName();
                Object o=elements.get(name);
                if (null!=o)
                {
                    try
                    {
                        String dname=(String)o;
                        if (dname.endsWith("()"))
                        {
                            if (callFunction(destination,dname,funcData,node))
                            {
                                found.put(dname,dname);
                            }
                        }
                        else
                        {
                            String data=getTextData(node);
                            if(putVariable(destination,dname,data))
                            {
                                found.put(dname,dname);
                            }
                        }
                    }
                    catch (Exception ex)
                    {
                        ex.printStackTrace();
                        throw new Exception("error while trying to set "+name+": "+ex.getMessage());
                    }
                }
            }
        }
        // Regeln auswerten
        for (int i=0;i<checkRules.length;i++)
        {
            String rule=checkRules[i];
            if (!checkRule(rule,found))
            {
                throw new Exception("rule ("+rule+") not fulfilled");
            }
        }
    }
    
    /**
        reads a XML node, parses it and writes the values according to the rules
        in a destination object
        attributes in the destination object must be public

        rules:
        2-dimensional string array
        {{"var name","element name"}, // writes the content of this element into the variable 
         {"method name()","element name"}} // calls the method with the element
                                         // parameters are the block and a properties opbject (funcData)
               
        checking rules:
        set of strings, in RPN:
        "var1" // var1 must have assigned a value
        "var2 var3 ||" // var3 and var2 must have assigned a value
        "var4 var5 && var6 var7 && ||" // var4 and var5 OR var6 and var7 must have assigned a value
        var means a variable name or a method name, which get assigned a value (or was called)
        each rule must me true after a boolean evaluation
        operators: && || !
        
        @param configuration the node with the configuration data
        @param rules the mapping rules
        @param checkRules the checking rules
        @param funcData the data which is given to each called method
        @param destination the destination object
    */
    public static void readConfigElement(Block configuration,String[][] rules, String[] checkRules, Properties funcData, Object destination) throws Exception
    { 
        Hashtable elements=new Hashtable();
        Hashtable found=new Hashtable();
        for (int ruleNum=0;ruleNum<rules.length;ruleNum++)
        {
            try
            {
                String[] rule=rules[ruleNum];
                String dest=rule[0];
                String name=rule[1];
                elements.put(name,dest);
            }
            catch (StringIndexOutOfBoundsException ex)
            {
                ex.printStackTrace();
                throw new Exception("Rule "+ruleNum+" has not enough elements");
            }
        }
        // sub-nodes
        Enumeration childs=configuration.getAllChilds();
        while (childs.hasMoreElements())
        {
            Block block=(Block)childs.nextElement();
            String name=block.getName();
            Object o=elements.get(name);
            if (null!=o)
            {
                try
                {
                    String dname=(String)o;
                    if (dname.endsWith("()"))
                    {
                        if (callFunction(destination,dname,funcData,block))
                        {
                            found.put(dname,dname);
                        }
                    }
                    else
                    {
                        String data=block.getValue().getValue();
                        if(putVariable(destination,dname,data))
                        {
                            found.put(dname,dname);
                        }
                    }
                }
                catch (Exception ex)
                {
                    ex.printStackTrace();
                    throw new Exception("error while trying to set "+name+": "+ex.getMessage());
                }
            }

        }
        // Regeln auswerten
        for (int i=0;i<checkRules.length;i++)
        {
            String rule=checkRules[i];
            if (!checkRule(rule,found))
            {
                throw new Exception("rule ("+rule+") not fulfilled");
            }
        }
    }

    private static boolean callFunction(Object destination, String name, Properties funcData, Node data) throws Exception
    {
        name=name.substring(0,name.length()-2);
        try
        {
            Class dClass=destination.getClass(); 
            Class[] ca={Node.class,Properties.class};
            Method m=dClass.getDeclaredMethod(name,ca);
            if (null==m)
            {
                throw new Exception("can't find method");
            }
            Object[] oa={data,funcData};
            m.invoke(destination,oa);
        }
        catch (Exception ex)
        {
            throw new Exception("can't find method "+name);
        }
        return true;
    }
    
        private static boolean callFunction(Object destination, String name, Properties funcData, Block data) throws Exception
        {
            name=name.substring(0,name.length()-2);
            try
            {
                Class dClass=destination.getClass(); 
                Class[] ca={Block.class,Properties.class};
                Method m=dClass.getDeclaredMethod(name,ca);
                if (null==m)
                {
                    throw new Exception("can't find method");
                }
                Object[] oa={data,funcData};
                m.invoke(destination,oa);
            }
            catch (Exception ex)
            {
                throw new Exception("can't find method "+name);
            }
            return true;
        }

    private static boolean putVariable(Object destination, String name, String data) throws Exception
    {
        data=data.trim();
        Class dClass=destination.getClass();
        Field f=dClass.getDeclaredField(name);
        if (null==f)
        {
            throw new Exception("can't find field");
        }
        Class fClass=f.getType();
        String fType=fClass.getName();
        Object o=null;
        if (fType.equals("int"))
        {
            o=Integer.valueOf(data);
        }
        else if (fType.equals("long"))
        {
            o=Long.valueOf(data);
        }
        else if (fType.equals("short"))
        {
            o=Short.valueOf(data);
        }
        else if (fType.equals("float"))
        {
            o=Float.valueOf(data);
        }
        else if (fType.equals("double"))
        {
            o=Double.valueOf(data);
        }
        else if (fType.equals("boolean"))
        {
            data=data.toLowerCase(); 
            if (data.equals("1")||data.equals("yes")|data.equals("true"))
            {
                o=new Boolean(true);
            }
            else
            {
                o=new Boolean(false);
            }
        }
        else if (fType.equals("char"))
        {
            o=new Character((data.length()>0)?data.charAt(0):0);
        }
        else if (fType.equals("byte"))
        {
            o=Byte.valueOf(data);
        }
        else if (fType.equals("java.lang.String"))
        {
            o=data;
        }
        else
        {
            throw new Exception("unknown data type: "+fType);
        }
        f.set(destination,o);
        return true;
    }
    
    // evaluates a equation in RPN
    private static boolean checkRule(String rule, Hashtable found)
    {
        Stack stack=new Stack();
        while (true)
        {
            String element=null;
            rule=rule.trim();
            int pos=rule.indexOf(' ');
            if (-1==pos)
            {
                element=rule;
                rule="";
            }
            else
            {
                element=rule.substring(0,pos);
                rule=rule.substring(pos);
            }
            if (0==element.length())
                break;
            if (Character.isJavaIdentifierStart(element.charAt(0)))
            {
                Object o=found.get(element);
                if (null==o)
                {
                    stack.push("0");
                }
                else
                {
                    stack.push("1");
                }
            }
            else if (element.equals("&&"))
            {
                boolean b1=string2bool((String)stack.pop());
                boolean b2=string2bool((String)stack.pop());
                stack.push((b1&&b2)?"1":"0");
            }
            else if (element.equals("||"))
            {
                boolean b1=string2bool((String)stack.pop());
                boolean b2=string2bool((String)stack.pop());
                stack.push((b1||b2)?"1":"0");
            }
            else if (element.equals("!"))
            {
                boolean b1=string2bool((String)stack.pop());
                stack.push((!b1)?"1":"0");
            }
        }
        return string2bool((String)stack.pop());
    }
    
    private static boolean string2bool(String s)
    {
        if (s.equals("1"))
            return true;
        return false;
    }
}
